/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.bsvrz.dua.bastband.bastband.dvs.funclib;

import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Set;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.Dataset;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.bastband.bastband.dvs.kex.datenaustausch.output.ChannelException;
import de.bsvrz.dua.bastband.bastband.dvs.kex.datenaustausch.output.DataProviderException;

/**
 * <p>
 * Klasse zum Lesen von Daten f&uuml;r eine Reihe von Zeitpunkten innerhalb
 * eines Zeitraumes aus dem Archiv und die weitere Verarbeitung durch einen oder
 * mehrere {@link ResultDataChannel}(s).
 * </p>
 *
 * <p>
 * In der Methode {@link #processData()} wird f&uuml;r alle Zeitpunkte die
 * aktuell g&uuml;tigen Daten der Kombinationen SystemObject/DataDescription
 * eines Zeitpunktes als ResultData[] an den oder die {@link ResultDataChannel
 * ResultDataChannel(s)} &uuml;bergeben.
 * </p>
 *
 * <p>
 * Die zu verarbeitenden Kombinationen aus SystemObject und DataDescription
 * werden durch die Methode
 * {@link #addDataCombination(SystemObject so, DataDescription dataDescription)
 * addDataKombination} festgelegt.
 * </p>
 * Die Reihenfolge im an den/die Channel &uuml;bergebenen DataResult[]
 * entspricht der Reihenfolge der Methodenaufrufe
 * {@link #addDataCombination(SystemObject so, DataDescription dataDescription)}
 * .
 *
 * <pre>
 *   Die Zeitpunkte wird bestimmt durch:
 *     1. Zeitraum (Period) - Periodenanfang, Periodenl&auml;nge, Einheit der Periodenl&auml;nge
 *     2. Schrittweite (Increment) - IncrementWert, IncrementEinheit
 * </pre>
 *
 * <p>
 * Der erste Zeitpunkt ist gleich dem Periodenanfang, der letzte ist kleiner als
 * Periodenanfang + Periodenl&auml;nge.
 *
 * <pre>
 *  Beispiel f&uuml;r die Berechnung der Zeitpunkte:
 *  <code>
 *      ResultDataProvider provider = new ResultDataProvider();
 *      provider.setIncrement(1, Calendar.HOUR);
 *      Calendar cal = Calendar.getInstance();
 *      cal.set(2004, Calendar.JANUARY, 1, 0, 0, 0);
 *      cal.set(Calendar.MILLISECOND, 0);
 *      provider.setPeriod(cal, 5, Calendar.HOUR);
 *      provider.checkSequence(System.out);
 *  </code>
 *  ergibt:
 *      01.01.2004 00:00:00.000
 *      01.01.2004 01:00:00.000
 *      01.01.2004 02:00:00.000
 *      01.01.2004 03:00:00.000
 *      01.01.2004 04:00:00.000
 * </pre>
 *
 * <p>
 * Die Daten werden nicht unbedingt f&uuml;r den gesamten zu bearbeitenen
 * Zeitraum aus der Datenbank geholt, sondern abschnittsweise. Die L&auml;nge
 * des Zeitbereiches, f&uuml;r den die Daten mit einmal aus dem Archiv gelesen
 * wird, ergibt sich aus (ReadIntervallLength * IncrementUnit).
 * </p>
 *
 * @version 1.0
 */
public class ResultDataProvider extends java.lang.Object {

	private static final de.bsvrz.dua.bastband.bastband.funclib.logger.Logger LOGGER = de.bsvrz.dua.bastband.bastband.funclib.logger.LoggerFactory
			.getLoggerFactory().getLogger();

	private static int ArsType_UNASSIGNED = 0;
	private static int ArsType_LOGOS = 1;
	private static int ArsType_MARZ = 2;

	private static SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss.SSS ");

	ArrayList<SystemObject> _ListSystemObject = new ArrayList<>(6);
	ArrayList<DataDescription> _DataDescriptionList = new ArrayList<>(6);

	Calendar _PeriodStart;

	int __ArsType = ResultDataProvider.ArsType_UNASSIGNED;

	/**
	 * wird zwischendurch mal auf den gerade bearbeiteten Zeitpunkt gesetzt um einen
	 * Fortschritt zu ermitteln.
	 */
	Calendar _ProcessingCalendar = Calendar.getInstance();

	boolean _IsNotInterrupted = true;

	int _PeriodLength = 1;
	int _PeriodLengthUnit;

	int _IncrementValue = 1;
	int _IncrementUnit;

	int _ReadIntervallLength = 24;

	List<ResultDataChannel> _DataChannelList = new ArrayList<>();

	public ResultDataProvider() {
		if (__ArsType == ResultDataProvider.ArsType_UNASSIGNED) {
			try {
				LogosClientDavConnection.getInstance().getArchive();
				__ArsType = ResultDataProvider.ArsType_MARZ;
			} catch (@SuppressWarnings("unused") final Exception ex) {
				__ArsType = ResultDataProvider.ArsType_LOGOS;
			}
		}
	}

	/**
	 * setzt den Zeitraum f&uuml;r den die Zeitpunkte erzeugt werden. Der erste
	 * Zeitpunkt ist gleich dem Anfangszeitpunkt.
	 *
	 * @param periodStart
	 *            Anfang der Zeitraumes
	 * @param periodLength
	 *            Wert der L&auml;nge der Zeitraumes.
	 * @param unit
	 *            Einheit der Zeitrauml&auml;nge
	 */
	public void setPeriod(final Calendar periodStart, final int periodLength, final int unit) {
		this._PeriodStart = periodStart;
		this._PeriodLength = periodLength;
		this._PeriodLengthUnit = unit;
	}

	/**
	 * setzt die Schrittweite zur Berechnung der Zeitpunkte. Der erste Zeitpunkt ist
	 * gleich dem Anfangszeitpunktes des Zeitraumes.
	 *
	 * @param incrementValue
	 *            Wert der Schrittweite
	 * @param incrementUnit
	 *            Einhet der Schrittweite
	 */
	public void setIncrement(final int incrementValue, final int incrementUnit) {
		this._IncrementValue = incrementValue;
		this._IncrementUnit = incrementUnit;
	}

	/**
	 * berechnet den gesamten zu bearbeitenden Zeitraum in Minuten, um ein Mass
	 * f&uuml;r den Fortschritt zu haben.
	 * 
	 * @param calPeriodStart
	 *            der Anfangszeitpunkt der Periode
	 * @param periodLengthUnit
	 *            die Einheit der Länge
	 * @param periodLength
	 *            die Länge
	 *
	 * @return der insgesamt zu bearbeitende Zeitraum.
	 */
	public static int calculateTotalPeriodInMinutes(final Calendar calPeriodStart, final int periodLengthUnit,
			final int periodLength) {
		final Calendar calEnd = (Calendar) calPeriodStart.clone();
		calEnd.add(periodLengthUnit, periodLength);
		final long lTimeDiffInMillis = calEnd.getTimeInMillis() - calPeriodStart.getTimeInMillis();
		return (int) (lTimeDiffInMillis / 1000 / 60);
	}

	/**
	 * berechnet den gesamten zu bearbeitenden Zeitraum in Minuten, um ein Mass
	 * f&uuml;r den Fortschritt zu haben.
	 * 
	 * @param calPeriodStart
	 *            der Anfangszeitpunkt der Periode
	 * @param calPeriodEnd
	 *            der Endzeitpunkt der Periode
	 *
	 * @return der insgesamt zu bearbeitende Zeitraum.
	 */
	public static int calculateTotalPeriodInMinutes(final Calendar calPeriodStart, final Calendar calPeriodEnd) {
		final long lTimeDiffInMillis = calPeriodEnd.getTimeInMillis() - calPeriodStart.getTimeInMillis();
		return (int) (lTimeDiffInMillis / 1000 / 60);
	}

	/**
	 * gibt den bereits bearbeiteten Zeitraum in Minuten zur&uuml;ck, um ein Mass
	 * f&uuml;r den Fortschritt zu haben.
	 *
	 * @return der insgesamt zu bearbeitende Zeitraum.
	 */
	public int getProgressedPeriodInMinutes() {
		final long lTimeDiffInMillis = _ProcessingCalendar.getTimeInMillis() - _PeriodStart.getTimeInMillis();
		return (int) (lTimeDiffInMillis / 1000 / 60);
	}

	/**
	 * meldet einen Channel an.
	 *
	 * @param dataChannel
	 *            ResultDataChannel
	 */
	public void addChannel(final ResultDataChannel dataChannel) {
		this._DataChannelList.add(dataChannel);
	}

	/**
	 * l&ouml;scht alle Channel.
	 */
	public void clearChannels() {
		this._DataChannelList.clear();
	}

	/**
	 * f&uuml;gt eine Kombination aus SystemObject und DataDescription zur Liste der
	 * Kombinationen hinzu, dessen Daten verarbeitet werden soll. Die Daten der
	 * Kombination werden entsprechend der Reihenfolge in der Liste zum ResultData[]
	 * zusammengefasst und an die Channels &uuml;bergeben.
	 *
	 * @param so
	 *            SystemObject
	 * @param dataDescription
	 *            DataDescription
	 */
	public void addDataCombination(final SystemObject so, final DataDescription dataDescription) {
		this._ListSystemObject.add(so);
		this._DataDescriptionList.add(dataDescription);
	}

	/**
	 * l&ouml;scht alle DatenKombinationen.
	 */
	public void clearDataCombinations() {
		this._ListSystemObject.clear();
		this._DataDescriptionList.clear();
	}

	/**
	 * f&uuml;r alle Zeitpunkte werden die Daten der Kombinationen aus SystemObject
	 * und DataDescription aus dem Archiv gelesen und an die die angemeldeten
	 * Channels &uuml;bergeben.
	 *
	 * @throws DataNotSubscribedException
	 *             fehlende Datenanmeldung
	 * @throws ChannelException
	 *             Fehler beim Verarbeiten der Daten
	 * @throws DataProviderException
	 *             Datenexport war nicht erfolgreich
	 */
	public void processData() throws ChannelException, DataProviderException {
		final LogosClientDavConnection davConn = LogosClientDavConnection.getInstance();

		final int soCount = _ListSystemObject.size();

		final SystemObject[] soArray = new SystemObject[1];
		final ArchiveData[][] rsArrays = new ArchiveData[soCount][];

		final Calendar endCalendar = (Calendar) _PeriodStart.clone();
		endCalendar.add(_PeriodLengthUnit, _PeriodLength);
		_ProcessingCalendar.setTimeInMillis(_PeriodStart.getTimeInMillis());

		final Calendar rollingStartCalendar = (Calendar) _PeriodStart.clone();
		Calendar rollingEndCalendar = (Calendar) _PeriodStart.clone();

		try {
			while ((rollingEndCalendar.before(endCalendar)) && _IsNotInterrupted) {
				rollingEndCalendar.add(_IncrementUnit, _ReadIntervallLength);
				if (rollingEndCalendar.after(endCalendar)) {
					rollingEndCalendar = endCalendar;
				}
				for (int i = 0; i < soCount; i++) {
					final String soPid = _ListSystemObject.get(i).getPid();
					ArchiveData[] rsDataArray = null;
					ResultDataProvider.LOGGER
							.finer("Anfrage: " + ResultDataProvider.DATE_FORMAT.format(rollingStartCalendar.getTime())
									+ "  " + ResultDataProvider.DATE_FORMAT.format(rollingEndCalendar.getTime()));
					try {
						final DataDescription dataDescr = _DataDescriptionList.get(i);
						final String atgPid = dataDescr.getAttributeGroup().getPid();
						final String aspPid = dataDescr.getAspect().getPid();
						ResultDataProvider.LOGGER.finer("Anfrage von/bis="
								+ ResultDataProvider.DATE_FORMAT
										.format(new java.util.Date(_PeriodStart.getTimeInMillis()))
								+ "  " + ResultDataProvider.DATE_FORMAT
										.format(new java.util.Date(endCalendar.getTimeInMillis())));
						final Set<ArchiveData> arsData = ArchiveRequestHelper.getArchivData(davConn, soPid, atgPid,
								aspPid, rollingStartCalendar.getTimeInMillis(),
								rollingEndCalendar.getTimeInMillis() - 10);

						rsDataArray = arsData.toArray(new ArchiveData[arsData.size()]);

					} catch (final Exception e) {
						final DataDescription dataDescr = _DataDescriptionList.get(i);
						e.printStackTrace();
						throw new DataProviderException(
								"Fehler beim Lesen " + soArray[0].getPid() + ':' + dataDescr.getAttributeGroup() + ':'
										+ dataDescr.getAspect() + ':' + dataDescr.getSimulationVariant(),
								e);
					}
					if ((rsDataArray != null) && (rsDataArray.length > 0)) {
						rsArrays[i] = rsDataArray;
					} else {
						rsArrays[i] = null;
					}
				}

				_ProcessingCalendar.setTimeInMillis(rollingEndCalendar.getTimeInMillis());
				for (; (rollingStartCalendar.before(rollingEndCalendar) && _IsNotInterrupted); rollingStartCalendar
						.add(_IncrementUnit, _IncrementValue)) {

					final Dataset[] rsArrayAtTime = new Dataset[soCount];
					for (int i = 0; i < soCount; i++) {
						rsArrayAtTime[i] = getValidResultData(rsArrays[i], rollingStartCalendar);
					}
					if ((rsArrayAtTime[0] != null)
							&& (rsArrayAtTime[0].getDataTime() != rollingStartCalendar.getTimeInMillis())) {
						ResultDataProvider.LOGGER.error(
								"Zeitstempel der Daten stimmt nicht mit dem übergebenen Calendar-Object überein.");
						throw new IllegalStateException(
								"Zeitstempel der Daten stimmt nicht mit dem übergebenen Calendar-Object überein.");
					}
					for (int channelNr = 0; channelNr < _DataChannelList.size(); channelNr++) {
						_DataChannelList.get(channelNr).processRecord(rsArrayAtTime, rollingStartCalendar);
					}
				}
				_ProcessingCalendar.setTimeInMillis(rollingEndCalendar.getTimeInMillis());
			}
		} catch (final DataNotSubscribedException ex) {
			throw new DataProviderException("Datenhaltung ist nicht angemeldet.", ex);
		}
	}

	public void interrupt() {
		_IsNotInterrupted = false;
	}

	/**
	 * erm&ouml;glicht das Pr&uuml;fen der Zeitpunkt, die sich aus der Periode und
	 * Schrittweise ergeben.
	 *
	 * @param out
	 *            PrintStream
	 */
	public void checkSequence(final PrintStream out) {
		final Calendar endCalendar = (Calendar) _PeriodStart.clone();
		endCalendar.add(_PeriodLengthUnit, _PeriodLength);

		final Calendar rollingStartCalendar = (Calendar) _PeriodStart.clone();
		Calendar rollingEndCalendar = (Calendar) _PeriodStart.clone();
		ResultDataProvider.LOGGER.finer("startCal: %s",
				ResultDataProvider.DATE_FORMAT.format(rollingStartCalendar.getTime()));
		ResultDataProvider.LOGGER.finer("endCal:   %s",
				ResultDataProvider.DATE_FORMAT.format(rollingEndCalendar.getTime()));

		while (rollingEndCalendar.before(endCalendar)) {
			rollingEndCalendar.add(_IncrementUnit, _ReadIntervallLength);
			ResultDataProvider.LOGGER
					.finer("endCal: " + ResultDataProvider.DATE_FORMAT.format(rollingEndCalendar.getTime()));
			if (rollingEndCalendar.after(endCalendar)) {
				rollingEndCalendar = endCalendar;
			}
			for (; rollingStartCalendar.before(rollingEndCalendar); rollingStartCalendar.add(_IncrementUnit,
					_IncrementValue)) {
				out.println("   " + ResultDataProvider.DATE_FORMAT.format(rollingStartCalendar.getTime()));
			}
		}
	}

	private Dataset getValidResultData(final Dataset[] rsArray, final Calendar calendar) {

		Dataset retValue = null;

		if (rsArray != null) {
			for (int i = rsArray.length - 1; ((retValue == null) && (i >= 0)); i--) {
				if ((calendar.getTimeInMillis() == rsArray[i].getDataTime()) && (rsArray[i].getData() != null)) {
					retValue = rsArray[i];
				}
			}
		}
		return retValue;
	}

	@Override
	public String toString() {
		final StringBuffer sb = new StringBuffer();
		sb.append("ResultDataProvider: (PeriodStart: ");
		sb.append(ResultDataProvider.DATE_FORMAT.format(_PeriodStart.getTime()));
		sb.append(" - PeriodenLaenge: ").append(_PeriodLength).append(' ')
				.append(ResultDataProvider.getTimeUnit(_PeriodLengthUnit));
		sb.append(" - Increment: ").append(_IncrementValue).append(' ')
				.append(ResultDataProvider.getTimeUnit(_IncrementUnit));
		sb.append(')');
		return sb.toString();
	}

	/**
	 * <p>
	 * setzt die ReadIntervallLength. (ReadIntervallLength * IncrementUnit) ergibt
	 * die Zeitspanne, f&uuml;r die Daten vom Archiv geholt werden.
	 * </p>
	 *
	 * @param readIntervallLength
	 *            int
	 */
	public void setReadIntervallLength(final int readIntervallLength) {
		this._ReadIntervallLength = readIntervallLength;
	}

	public final static String getTimeUnit(final int unit) {
		switch (unit) {
		case 1: {
			return "Year";
		}
		case 2: {
			return "Month";
		}
		case 5: {
			return "Day";
		}
		case 10: {
			return "Hour";
		}
		case 11: {
			return "HourOfDay";
		}
		case 12: {
			return "Minute";
		}
		case 13: {
			return "Second";
		}
		case 14: {
			return "Millisecond";
		}
		case 6: {
			return "DayOfYear";
		}
		case 7: {
			return "DayOfWeek";
		}
		case 8: {
			return "DayOfWeekInMonth";
		}
		case 4: {
			return "WeekOfMonth";
		}
		case 3: {
			return "WeekOfYear";
		}
		default: {
			return "unbekannt";
		}
		}
	}

}
