/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.bsvrz.dua.bastband.bastband.dvs.kex.datenaustausch.output;

/**
 * History:
 *
 * $Log: FileDataChannel.java,v $
 * Revision 1.6  2007/08/22 08:49:07  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.5  2007/07/26 08:01:42  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.4  2007/07/25 13:02:06  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.3  2007/07/04 11:48:42  Trier
 * Neuordnung
 *
 * Revision 1.2  2007/06/20 12:36:12  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.1.1.1  2007/06/20 09:56:27  Trier
 * dvs alles
 *
 * Revision 1.1  2005/01/21 14:40:11  Trier
 * <Kein Kommentar eingegeben>
 *
 */
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

/**
 *
 * <p>
 * implementiert die Grundfunktionalit&auml;t f&uuml;r einen
 * {@link de.bsvrz.dua.bastband.bastband.dvs.funclib.ResultDataChannel
 * ResultDataChannel}, der in eine Datei schreibt.
 * </p>
 * <p>
 * Durch {@link #reset(String fileName) reset} kann der Channel wieder verwendet
 * werden.
 * </p>
 *
 * @version 1.0
 */
abstract public class FileDataChannel implements de.bsvrz.dua.bastband.bastband.dvs.funclib.ResultDataChannel {

	private String _FileName;
	private PrintWriter _Writer;

	private final List<Object> _LineBuffer = new ArrayList<>();

	private boolean _IsOpen = false;

	/**
	 * erzeugt ein FileDataChannel. Zum Beschreiben des Channel muss mit
	 * {@link #reset(String fileName) reset} der Filename gesetzt werden.
	 */
	public FileDataChannel() {
	}

	/**
	 * erzeugt einen FileDataChannel, dessen Ausgabe in die Datei mit dem
	 * angegebenen Filenamen erfolgt.
	 *
	 * @param fileName
	 *            Name der zu beschreibenen Datei
	 */
	public FileDataChannel(final String fileName) {
		_FileName = fileName;
		_IsOpen = true;
	}

	/**
	 * ist der Channel ge&ouml;ffnet, werden gepufferte Daten in die Datei
	 * geschrieben und danach die Datei mit dem angegebenen Namen zur Ausgabe
	 * benutzt.
	 *
	 * @param fileName
	 *            String
	 * @throws ChannelException
	 *             der Datenkanal konnte nicht zurückgesetzt werden
	 */
	public void reset(final String fileName) throws ChannelException {
		if (_IsOpen) {
			close();
		}
		_FileName = fileName;
		_IsOpen = true;
	}

	/**
	 * schlie&szlig;t den Channel nachdem gepufferte Daten geschrieben wurden.
	 *
	 * @throws ChannelException
	 *             die Daten konnten nicht erfolgreich geschrieben werden
	 */
	@Override
	public void close() throws ChannelException {
		if (_IsOpen) {
			flush();
			if (_Writer != null) {
				_Writer.close();
				_Writer = null;
			}
			_LineBuffer.clear();
			_FileName = null;
			_IsOpen = false;
		}
	}

	/**
	 * schreibt gepufferte Daten in die Datei.
	 *
	 * @throws ChannelException
	 *             wenn beim Schreiben Fehler auftreten.
	 */
	synchronized public void flush() throws ChannelException {
		if (_IsOpen) {
			try {
				if (_Writer == null) {
					_Writer = new PrintWriter(new FileWriter(_FileName));
					writeHeader(_Writer);
				}
				while (_LineBuffer.size() > 0) {
					_Writer.println(_LineBuffer.remove(0));
				}
			} catch (final IOException e) {
				throw new ChannelException("IOException beim Schreiben.", e);
			}
		} else {
			throw new ChannelException("Channel is closed");
		}
	}

	/**
	 * erm&ouml;glicht das Schreiben von Daten am Anfang der Datei.
	 *
	 * @param writer
	 *            PrintWriter
	 * @throws IOException
	 *             wenn beim Schreiben Fehler auftreten.
	 */
	abstract protected void writeHeader(PrintWriter writer) throws IOException;

	/**
	 * schreibt Daten in den Puffer.
	 *
	 * @param o
	 *            zu schreibenes Object.
	 * @throws ChannelException
	 *             wenn der Channel geschlossen ist.
	 */
	synchronized public void spool(final Object o) throws ChannelException {
		if (_IsOpen) {
			_LineBuffer.add(o);
		} else {
			throw new ChannelException("Channel is closed");
		}
	}

	/**
	 * l&ouml;scht die letzte Zeile aus dem Puffer.
	 *
	 * @throws ChannelException
	 *             wenn der Channel geschlossen ist oder keine Zeilen im Puffer.
	 */
	synchronized public void deleteLastLine() throws ChannelException {
		if ((_IsOpen) && (_LineBuffer.size() > 0)) {
			_LineBuffer.remove(_LineBuffer.size() - 1);
		} else {
			throw new ChannelException("Channel is closed");
		}
	}

}
